from ford.pagetree import get_page_tree
from ford._markdown import MetaMarkdown

from textwrap import dedent
from locale import getpreferredencoding


def test_footnotes_on_one_page(tmp_path):
    """Check that footnotes only appear on the pages they are defined on,
    issue #327"""

    with open(tmp_path / "index.md", "w") as f:
        f.write("title: Index")

    with open(tmp_path / "a.md", "w") as f:
        f.write(
            dedent(
                """\
            title: Page A

            This has a footnote[^1] that should only appear on this page

            [^1]: This is the footnote on page A
            """
            )
        )

    with open(tmp_path / "b.md", "w") as f:
        f.write(
            dedent(
                """\
            title: Page B

            This page should not have any footnotes
            """
            )
        )

    md = MetaMarkdown()
    result_dir = tmp_path / "result"
    nodes = get_page_tree(tmp_path, result_dir, tmp_path / "doc", md)

    assert len(nodes.subpages) == 2
    assert "This is the footnote on page A" in nodes.subpages[0].contents
    assert "This is the footnote on page A" not in nodes.subpages[1].contents


def test_footnotes_on_one_page_parse_failure(tmp_path):
    """Check that footnotes only appear on the pages they are defined on,
    even if we fail to parse one of the files, issue #327"""

    with open(tmp_path / "index.md", "w") as f:
        f.write("title: Index")

    with open(tmp_path / "a.md", "w") as f:
        f.write(
            dedent(
                """\
            This page is missing a title and so expected not to be parsed

            This has a footnote[^1] that should only appear on this page

            [^1]: This is the footnote on page A
            """
            )
        )

    with open(tmp_path / "b.md", "w") as f:
        f.write(
            dedent(
                """\
            title: Page B

            This page should not have any footnotes
            """
            )
        )

    md = MetaMarkdown()
    result_dir = tmp_path / "result"
    nodes = get_page_tree(tmp_path, result_dir, tmp_path / "doc", md)

    assert len(nodes.subpages) == 1
    assert "This is the footnote on page A" not in nodes.subpages[0].contents


def test_non_utf8_encoding(tmp_path):
    """This is not really a full test for issue #518, as it only tests
    the lower part of the call-tree. A more thorough test would do
    something like encode a whole project in the non-default encoding.

    """
    # Try to get an encoding which is *not* the default for `open`
    encoding = "gbk" if getpreferredencoding().lower() == "utf-8" else "utf-8"

    with open(tmp_path / "index.md", "wb") as f:
        f.write(
            dedent(
                """\
        ---
        title: Specification
        ---

        @warning
        本文档是一个简易的规范文档，仅供参考
        """
            ).encode(encoding)
        )

    md = MetaMarkdown()
    result_dir = tmp_path / "result"
    nodes = get_page_tree(tmp_path, result_dir, tmp_path / "doc", md, encoding=encoding)

    assert "本文档是" in nodes.contents



from ford.settings import (
    is_same_type,
    ProjectSettings,
    load_markdown_settings,
    EntitySettings,
    ExtraFileType,
)

from typing import List, Optional
from textwrap import dedent
import pytest

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib  # type: ignore[no-redef]


@pytest.mark.parametrize(
    ("type_in", "tp", "expected"),
    (
        (list, list, True),
        (list, str, False),
        (List[str], List[str], True),
        (Optional[str], str, True),
    ),
)
def test_is_same_type(type_in, tp, expected):
    assert is_same_type(type_in, tp) is expected


def test_settings_type_conversion():
    settings = ProjectSettings(src_dir="./src")

    assert settings.src_dir == ["./src"]


def test_settings_type_conversion_from_markdown():
    settings, _ = load_markdown_settings(
        ".",
        dedent(
            """\
            ---
            project: some project
            src_dir: source
            summary: first
                     second
            preprocess: true
            fpp_extensions: fpp
                            F90
            max_frontpage_items: 4
            alias: a = b
                c = d
            extra_filetypes: cpp //
                sh # bash
                py # python
            ---
            """
        ),
    )

    assert settings.src_dir == ["source"]
    assert settings.fpp_extensions == ["fpp", "F90"]
    assert settings.project == "some project"
    assert settings.summary == "first\nsecond"
    assert settings.preprocess is True
    assert settings.max_frontpage_items == 4
    assert settings.alias == {
        "a": "b",
        "c": "d",
    }
    assert settings.extra_filetypes == {
        "cpp": ExtraFileType("cpp", "//"),
        "sh": ExtraFileType("sh", "#", "bash"),
        "py": ExtraFileType("py", "#", "python"),
    }


def test_settings_from_toml():
    text = dedent(
        '''\
    project = "some project"
    src_dir = "source"
    summary = """
    first
    second"""
    preprocess = true
    fpp_extensions = ["fpp", "F90"]
    max_frontpage_items = 4
    alias = {a = "b", c = "d"}
    extra_filetypes = [
      { extension = "cpp", comment = "//" },
      { extension = "sh", comment = "#", lexer = "bash" },
      { extension = "py", comment = "#", lexer = "python" },
    ]
    '''
    )

    settings = ProjectSettings(**tomllib.loads(text))

    assert settings.src_dir == ["source"]
    assert settings.fpp_extensions == ["fpp", "F90"]
    assert settings.project == "some project"
    assert settings.summary == "first\nsecond"
    assert settings.preprocess is True
    assert settings.max_frontpage_items == 4
    assert settings.alias == {
        "a": "b",
        "c": "d",
    }
    assert settings.extra_filetypes == {
        "cpp": ExtraFileType("cpp", "//"),
        "sh": ExtraFileType("sh", "#", "bash"),
        "py": ExtraFileType("py", "#", "python"),
    }


def test_entity_settings_from_project():
    project_settings = ProjectSettings(source=True)
    entity_settings = EntitySettings.from_project_settings(project_settings)
    assert entity_settings.source is True


def test_update_entity_settings():
    expected_version = "1.0.1"
    settings = EntitySettings(source=True)
    settings.update({"version": expected_version})

    assert settings.source is True
    assert settings.version == expected_version


def test_extra_filetype():
    c = ExtraFileType.from_string("c //")
    assert c == ExtraFileType("c", "//")

    c_with_lexer = ExtraFileType.from_string("c // c_lexer")
    assert c_with_lexer == ExtraFileType("c", "//", "c_lexer")


def test_extra_filetype_error():
    with pytest.raises(ValueError):
        ExtraFileType.from_string("c")

    with pytest.raises(ValueError):
        ExtraFileType.from_string("c // c lexer")



#!/usr/bin/env python
# -- coding: utf-8 --
#
#  ford.py
#
#  Copyright 2014 Christopher MacMackin <cmacmackin@gmail.com>
#  This file is part of FORD.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

from contextlib import contextmanager
import copy
from io import StringIO
import sys
import argparse
import os
import pathlib
import subprocess
from datetime import datetime
import time
from typing import Tuple, Optional
from textwrap import dedent

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib  # type: ignore[no-redef]

from ford.external_project import dump_modules
import ford.fortran_project
import ford.sourceform
import ford.output
from ford.utils import ProgressBar
from ford._typing import PathLike
from ford.pagetree import get_page_tree
from ford._markdown import MetaMarkdown
from ford.version import __version__
from ford.settings import (
    ProjectSettings,
    load_markdown_settings,
    load_toml_settings,
    convert_types_from_commandarguments,
)


__appname__ = "FORD"
__author__ = "Chris MacMackin"
__credits__ = [
    "Balint Aradi",
    "Iain Barrass",
    "Izaak Beekman",
    "Jérémie Burgalat",
    "David Dickinson",
    "Gavin Huttley",
    "Harald Klimach",
    "Nick R. Papior",
    "Marco Restelli",
    "Schildkroete23",
    "Stephen J. Turnbull",
    "Jacob Williams",
    "Stefano Zhagi",
]
__license__ = "GPLv3"
__maintainer__ = "Chris MacMackin"
__status__ = "Production"


@contextmanager
def stdout_redirector(stream):
    old_stdout = sys.stdout
    sys.stdout = stream
    try:
        yield
    finally:
        sys.stdout = old_stdout


LICENSES = {
    "by": (
        '<a rel="license" href="http://creativecommons.org/licenses/by/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by/4.0/80x15.png" /></a>'
    ),
    "by-nd": (
        '<a rel="license" href="http://creativecommons.org/licenses/by-nd/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-nd/4.0/80x15.png" /></a>'
    ),
    "by-sa": (
        '<a rel="license" href="http://creativecommons.org/licenses/by-sa/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-sa/4.0/80x15.png" /></a>'
    ),
    "by-nc": (
        '<a rel="license" href="http://creativecommons.org/licenses/by-nc/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-nc/4.0/80x15.png" /></a>'
    ),
    "by-nc-nd": (
        '<a rel="license" href="http://creativecommons.org/licenses/by-nc-nd/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-nc-nd/4.0/80x15.png" /></a>'
    ),
    "by-nc-sa": (
        '<a rel="license" href="http://creativecommons.org/licenses/by-nc-sa/4.0/">'
        '<img alt="Creative Commons License" style="border-width:0" src="https://i.creativecommons.org/l/by-nc-sa/4.0/80x15.png" /></a>'
    ),
    "gfdl": '<a rel="license" href="http://www.gnu.org/licenses/old-licenses/fdl-1.2.en.html">GNU Free Documentation License</a>',
    "opl": '<a rel="license" href="http://opencontent.org/openpub/">Open Publication License</a>',
    "pdl": '<a rel="license" href="http://www.openoffice.org/licenses/PDL.html">Public Documentation License</a>',
    "bsd": '<a rel="license" href="http://www.freebsd.org/copyright/freebsd-doc-license.html">FreeBSD Documentation License</a>',
    "isc": '<a rel="license" href="https://opensource.org/licenses/ISC">ISC (Internet Systems Consortium) License</a>',
    "mit": '<a rel="license" href="https://opensource.org/licenses/MIT">MIT</a>',
    "": "",
}


def initialize():
    """
    Method to parse and check configurations of FORD, get the project's
    global documentation, and create the Markdown reader.
    """
    args = get_command_line_arguments()

    # Read in the project-file. This will contain global documentation (which
    # will appear on the homepage) as well as any information about the project
    # and settings for generating the documentation

    proj_docs = args.project_file.read()
    directory = os.path.dirname(args.project_file.name)
    proj_docs, proj_data = load_settings(proj_docs, directory, args.project_file.name)

    return parse_arguments(vars(args), proj_docs, proj_data, directory)


def get_command_line_arguments() -> argparse.Namespace:
    """Read the command line arguments"""

    parser = argparse.ArgumentParser(
        "ford",
        description="Document a program or library written in modern Fortran. "
        "Any command-line options over-ride those specified in the project file.",
    )
    parser.add_argument(
        "project_file",
        help="file containing the description and settings for the project",
        type=argparse.FileType("r", encoding="UTF-8"),
    )
    parser.add_argument(
        "-d",
        "--src_dir",
        action="append",
        help="directories containing all source files for the project (default: ``./src``)",
    )
    parser.add_argument(
        "-p",
        "--page_dir",
        help="directory containing static pages (default: None)",
    )
    parser.add_argument(
        "-o",
        "--output_dir",
        help="directory in which to place output files (default: ``./doc``)",
    )
    parser.add_argument("-s", "--css", help="custom style-sheet for the output")
    parser.add_argument(
        "-r",
        "--revision",
        dest="revision",
        help="source code version number or revision of the project to document",
    )
    parser.add_argument(
        "--exclude",
        action="append",
        help="list of files which should not be included in the documentation",
    )
    parser.add_argument(
        "--exclude_dir",
        action="append",
        help="list of directories whose contents should not be included in the documentation",
    )
    parser.add_argument(
        "-e",
        "--extensions",
        action="append",
        help="extensions which should be scanned for documentation (default: ``f90, f95, f03, f08``)",
    )
    parser.add_argument(
        "-m",
        "--macro",
        action="append",
        help="preprocessor macros (optionally with values) to be applied to preprocessed files",
    )
    parser.add_argument(
        "-w",
        "--warn",
        dest="warn",
        action="store_true",
        default=None,
        help="display warnings for undocumented items",
    )
    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        default=None,
        help="try to continue to read files even if there are fatal errors",
    )
    parser.add_argument(
        "-g",
        "--graph",
        dest="graph",
        action="store_true",
        default=None,
        help="generate graphs for documentation output",
    )
    parser.add_argument(
        "--no-search",
        dest="search",
        action="store_false",
        default=None,
        help="don't process documentation to produce a search feature",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_true",
        default=None,
        help="suppress normal output",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s version {__version__}",
    )
    parser.add_argument(
        "--debug",
        dest="dbg",
        action="store_true",
        default=None,
        help="display traceback if fatal exception occurs and print faulty line",
    )
    parser.add_argument(
        "-I",
        "--include",
        action="append",
        help="list of directories to be searched for ``include`` files",
    )
    parser.add_argument(
        "--externalize",
        action="store_true",
        default=None,
        help="provide information about Fortran objects in a json database for "
        "other FORD projects to refer to.",
    )
    parser.add_argument(
        "-L",
        "--external_links",
        dest="external",
        action="append",
        help="""external projects to link to.
        If an entity is not found in the sources, FORD will try to look it up in
        those external projects. If those have documentation generated by FORD with
        the externalize option, a link will be placed into the documentation wherever
        this entity is referred to.
        FORD will look in the provided paths for a modules.json file.
        """,
    )
    parser.add_argument(
        "--config",
        type=str,
        default=None,
        help="Any other FORD options as a semicolon-separated TOML string. "
        "Options set through this have lower precedence than other command line options",
    )

    return parser.parse_args()


def load_settings(
    proj_docs: str,
    directory: PathLike = pathlib.Path("."),
    filename: Optional[str] = None,
) -> Tuple[str, ProjectSettings]:
    """Load Ford settings from ``fpm.toml`` if present, or from
    metadata in supplied project file1

    Parameters
    ----------
    proj_docs : str
        Text of project file
    directory :
        Project directory

    Returns
    -------
    proj_docs: str
        Text of project file converted from markdown
    proj_data: dict
        Project settings

    """

    proj_data = load_toml_settings(directory)

    if proj_data is None:
        proj_data, proj_docs = load_markdown_settings(directory, proj_docs, filename)

    return proj_docs, proj_data


def parse_arguments(
    command_line_args: dict,
    proj_docs: str,
    proj_data: ProjectSettings,
    directory: PathLike = pathlib.Path("."),
):
    """Consolidates arguments from the command line and from the project
    file, and then normalises them how the rest of the code expects
    """

    if (config := command_line_args.get("config", None)) is not None:
        toml_string = "\n".join(config.split(";"))
        for key, value in tomllib.loads(toml_string).items():
            setattr(proj_data, key, value)

    # Get the default options, and any over-rides, straightened out
    proj_data = convert_types_from_commandarguments(proj_data, command_line_args)

    proj_data.normalise_paths(directory)

    proj_data.creation_date = (
        datetime.now().astimezone().strftime(proj_data.creation_date)
    )

    # Make sure no src_dir is contained within output_dir
    for srcdir in proj_data.src_dir:
        # In Python 3.9+ we can use pathlib.Path.is_relative_to
        if proj_data.output_dir in (srcdir, *srcdir.parents):
            raise ValueError(
                f"Source directory {srcdir} is a subdirectory of output directory {proj_data.output_dir}."
            )

    # Add gitter sidecar if specified in metadata
    if proj_data.gitter_sidecar is not None:
        proj_docs += f"""
        <script>
            ((window.gitter = {{}}).chat = {{}}).options = {{
            room: '{proj_data.gitter_sidecar.strip()}'
            }};
        </script>
        <script src="https://sidecar.gitter.im/dist/sidecar.v1.js" async defer></script>
        """

    # Handle preprocessor:
    if proj_data.preprocess:
        command = proj_data.preprocessor.split() + [os.devnull]
        # Check whether preprocessor works (reading nothing from stdin)
        try:
            subprocess.run(command, check=True, capture_output=True, text=True)
        except (subprocess.CalledProcessError, OSError) as ex:
            project_file = command_line_args["project_file"].name
            exit(
                dedent(
                    f"""\
                    Error: Testing preprocessor command (`{" ".join(command)}`) failed with error:
                        {ex.stderr.strip() if isinstance(ex, subprocess.CalledProcessError) else ex}

                    If you need to preprocess files, please fix the 'preprocessor' option in '{project_file}'.
                    Otherwise, please set 'preprocess: False' in '{project_file}'"""
                )
            )
    else:
        proj_data.fpp_extensions = []

    # Get the correct license for project license or use value as a custom license value.
    try:
        proj_data.license = LICENSES[proj_data.license.lower()]
    except KeyError:
        print(
            f'Notice: license "{proj_data.license}" is not a recognized value, using the value as a custom license value.'
        )
    # Get the correct license for doc license(website or doc) or use value as a custom license value.
    try:
        proj_data.doc_license = LICENSES[proj_data.doc_license.lower()]
    except KeyError:
        print(
            f'Notice: doc_license "{proj_data.doc_license}" is not a recognized value, using the value as a custom license value.'
        )

    return proj_data, proj_docs


def main(proj_data: ProjectSettings, proj_docs: str):
    """
    Main driver of FORD.
    """

    # Parse the files in your project
    project = ford.fortran_project.Project(proj_data)
    if len(project.files) < 1:
        print(
            "Error: No source files with appropriate extension found in specified directory."
        )
        sys.exit(1)

    print("  Correlating information from different parts of your project...")
    correlate_time_start = time.time()
    project.correlate()
    correlate_time_end = time.time()
    print(f"  ...done in {correlate_time_end - correlate_time_start:5.3f}s")

    # Setup Markdown object with any user-specified extensions
    aliases = copy.copy(proj_data.alias)
    aliases.update(proj_data.external)
    url_path = pathlib.Path(proj_data.project_url)
    aliases.update(
        {
            "url": str(url_path),
            "media": str(url_path / "media"),
            "page": str(url_path / "page"),
        }
    )

    md = MetaMarkdown(
        proj_data.md_base_dir,
        base_url=proj_data.project_url,
        extensions=proj_data.md_extensions,
        aliases=aliases,
        project=project,
    )

    # Convert the documentation from Markdown to HTML
    proj_docs = md.reset().convert(proj_docs, path=proj_data.project_url)
    project.markdown(md)

    # Convert summaries and descriptions to HTML
    if proj_data.summary is not None:
        proj_data.summary = md.convert(proj_data.summary)
    if proj_data.author_description is not None:
        proj_data.author_description = md.convert(proj_data.author_description)

    # Process any pages
    if proj_data.page_dir is not None:
        total_files = len(list(proj_data.page_dir.glob("**/*.md")))
        with ProgressBar("Processing pages", total=total_files) as progress:
            page_tree = get_page_tree(
                proj_data.page_dir,
                proj_data.copy_subdir,
                proj_data.output_dir,
                md,
                encoding=proj_data.encoding,
                progress=progress,
            )
    else:
        page_tree = None

    # Produce the documentation using Jinja2. Output it to the desired location
    # and copy any files that are needed (CSS, JS, images, fonts, source files,
    # etc.)

    docs = ford.output.Documentation(proj_data, proj_docs, project, page_tree)
    docs.writeout()

    if proj_data.externalize:
        # save FortranModules to a JSON file which then can be used
        # for external modules
        dump_modules(project, path=proj_data.output_dir)

    return 0


def run():
    proj_data, proj_docs = initialize()

    f = StringIO() if proj_data.quiet else sys.stdout
    with stdout_redirector(f):
        main(proj_data, proj_docs)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
#
#  utils.py
#  This file is part of FORD.
#
#  Copyright 2015 Christopher MacMackin <cmacmackin@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import re
import os
import os.path
import pathlib
from types import TracebackType
from typing import Dict, Union, List, Any, Tuple, Optional, Iterable, cast, Sized, Type
from io import StringIO
import itertools
from collections import defaultdict
from rich.progress import (
    Progress,
    BarColumn,
    TextColumn,
    SpinnerColumn,
    MofNCompleteColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
    TaskProgressColumn,
)

from ford.console import console


def get_parens(line: str, retlevel: int = 0, retblevel: int = 0) -> str:
    """
    By default takes a string starting with an open parenthesis and returns the portion
    of the string going to the corresponding close parenthesis. If retlevel != 0 then
    will return when that level (for parentheses) is reached. Same for retblevel.
    """
    if not line:
        return line
    parenstr = ""
    level = 0
    blevel = 0
    for char in line:
        if char == "(":
            level += 1
        elif char == ")":
            level -= 1
        elif char == "[":
            blevel += 1
        elif char == "]":
            blevel -= 1
        elif (
            (char.isalpha() or char in ("_", ":", ",", " "))
            and level == retlevel
            and blevel == retblevel
        ):
            return parenstr
        parenstr = parenstr + char

    if level == retlevel and blevel == retblevel:
        return parenstr
    raise RuntimeError(f"Couldn't parse parentheses: {line}")


def strip_paren(line: str, retlevel: int = 0) -> list:
    """
    Takes a string with parentheses and removes any of the contents inside or outside
    of the retlevel of parentheses. Additionally, whenever a scope of the retlevel is
    left, the string is split.

    e.g. strip_paren("foo(bar(quz) + faz) + baz(buz(cas))", 1) -> ["(bar() + faz)", "(buz())"]
    """
    retstrs = []
    curstr = StringIO()
    level = 0
    for char in line:
        if char == "(":
            if level == retlevel or level + 1 == retlevel:
                curstr.write(char)
            level += 1
        elif char == ")":
            if level == retlevel or level - 1 == retlevel:
                curstr.write(char)
            if level == retlevel:
                # We are leaving a scope of the desired level,
                # and should split to indicate as such.
                retstrs.append(curstr.getvalue())
                curstr = StringIO()
            level -= 1
        elif level == retlevel:
            curstr.write(char)

    if curstr.getvalue() != "":
        retstrs.append(curstr.getvalue())
    return retstrs


def paren_split(sep, string):
    """
    Splits the string into pieces divided by sep, when sep is outside of parentheses.
    """
    if len(sep) != 1:
        raise ValueError("Separation string must be one character long")
    retlist = []
    level = 0
    blevel = 0
    left = 0
    for i in range(len(string)):
        if string[i] == "(":
            level += 1
        elif string[i] == ")":
            level -= 1
        elif string[i] == "[":
            blevel += 1
        elif string[i] == "]":
            blevel -= 1
        elif string[i] == sep and level == 0 and blevel == 0:
            retlist.append(string[left:i])
            left = i + 1
    retlist.append(string[left:])
    return retlist


def quote_split(sep, string):
    """
    Splits the strings into pieces divided by sep, when sep in not inside quotes.
    """
    if len(sep) != 1:
        raise ValueError("Separation string must be one character long")
    retlist = []
    squote = False
    dquote = False
    left = 0
    i = 0
    while i < len(string):
        if string[i] == '"' and not dquote:
            if not squote:
                squote = True
            elif (i + 1) < len(string) and string[i + 1] == '"':
                i += 1
            else:
                squote = False
        elif string[i] == "'" and not squote:
            if not dquote:
                dquote = True
            elif (i + 1) < len(string) and string[i + 1] == "'":
                i += 1
            else:
                dquote = False
        elif string[i] == sep and not dquote and not squote:
            retlist.append(string[left:i])
            left = i + 1
        i += 1
    retlist.append(string[left:])
    return retlist


def str_to_bool(text):
    """Convert string to bool. Only takes 'true'/'false', ignoring case"""
    if isinstance(text, bool):
        return text
    if text.capitalize() == "True":
        return True
    if text.capitalize() == "False":
        return False
    raise ValueError(
        f"Could not convert string to bool: expected 'true'/'false', got '{text}'"
    )


def normalise_path(
    base_dir: pathlib.Path, path: Union[str, pathlib.Path]
) -> pathlib.Path:
    """Tidy up path, making it absolute, relative to base_dir"""
    return (base_dir / os.path.expandvars(path)).absolute().resolve()


def traverse(root, attrs) -> list:
    """Traverse a tree of objects, returning a list of all objects found
    within the attributes attrs"""
    nodes = []
    for obj in itertools.chain(*[getattr(root, attr, []) for attr in attrs]):
        nodes.append(obj)
        nodes.extend(traverse(obj, attrs))
    return nodes


# Global Vars
META_RE = re.compile(r"^[ ]{0,3}(?P<key>[A-Za-z0-9_-]+):\s*(?P<value>.*)")
META_MORE_RE = re.compile(r"^[ ]{4,}(?P<value>.*)")
BEGIN_RE = re.compile(r"^-{3}(\s.*)?")
END_RE = re.compile(r"^(-{3}|\.{3})(\s.*)?")


def meta_preprocessor(lines: Union[str, List[str]]) -> Tuple[Dict[str, Any], List[str]]:
    """Extract metadata from start of ``lines``

    This is modified from the `Meta-Data Python Markdown extension`_
    and uses the same syntax

    Arguments
    ---------
    lines:
        Text to process

    Returns
    -------
    meta:
        Dictionary of metadata, with lowercase keys
    lines:
        Original text with metadata lines removed as list

    Notes
    -----

    Original code copyright 2007-2008 `Waylan Limberg
    <http://achinghead.com>`_

    Changes copyright 2008-2014 `The Python Markdown Project
    <https://python-markdown.github.io>`_

    Further changes copyright 2023 Ford authors

    License: `BSD <https://opensource.org/licenses/bsd-license.php>`_

    .. _Meta-Data Python Markdown extension:
      https://python-markdown.github.io/extensions/meta_data/

    """

    if isinstance(lines, str):
        lines = lines.splitlines()

    if lines and BEGIN_RE.match(lines[0]):
        lines.pop(0)
    meta = defaultdict(list)
    key = None
    while lines:
        line = lines.pop(0)
        if line.strip() == "" or END_RE.match(line):
            break  # blank line or end of YAML header - done
        if m1 := META_RE.match(line):
            key = m1.group("key").lower().strip()
            value = m1.group("value").strip()
            meta[key].append(value)
        else:
            if (m2 := META_MORE_RE.match(line)) and key:
                # Add another line to existing key
                meta[key].append(m2.group("value").strip())
            else:
                lines.insert(0, line)
                break  # no meta data - done
    return meta, lines


class ProgressBar:
    """Progress bar that can be used to wrap an iterable and include
    the current item in the bar, or used as a context manager if not
    using a simple iterable

    An example::

        for item in (bar := ProgressBar("Processing items", items)):
            bar.set_current(item.name)
            item.do_work()

    Arguments
    ---------
    description:
        Description of task
    iterable:
        Collection to iterate over
    total:
        Can be used if length of iterable is not known at start

    """

    def __init__(
        self,
        description: str,
        iterable: Optional[Iterable] = None,
        total: Optional[int] = None,
    ):
        if total is None and hasattr(iterable, "__len__"):
            self._total: Optional[int] = len(cast(Sized, iterable))
        else:
            self._total = total

        # rich `Progress` breaks `pdb` and `breakpoint`, see:
        # - https://github.com/Textualize/rich/issues/1053
        # - https://github.com/Textualize/rich/issues/1465
        # and maintainer seems uninterested in fixing. So, workaround
        # by setting an env var to disable progress bar entirely
        disable: bool = bool(os.environ.get("FORD_DEBUGGING", False))

        self._iterable = iterable
        self._progress = Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description:21}"),
            BarColumn(),
            TaskProgressColumn(),
            MofNCompleteColumn(),
            TimeRemainingColumn(),
            TimeElapsedColumn(),
            TextColumn("{task.fields[current]}"),
            console=console,
            disable=disable,
        )

        self._progress.start()
        self._main_task = self._progress.add_task(
            description, total=self._total, current="--"
        )

    def set_current(self, current: str):
        """Set name of current item"""
        self._progress.update(self._main_task, advance=1, current=current)

    def __iter__(self):
        try:
            for item in self._iterable:
                yield item
        finally:
            self._progress.__exit__(None, None, None)

    def __enter__(self):
        self._progress.__enter__()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self._progress.__exit__(exc_type, exc_val, exc_tb)



from ford.fortran_project import Project, find_all_files
from ford.settings import ProjectSettings, ExtraFileType
from ford.utils import normalise_path
from ford.sourceform import (
    FortranVariable,
    FortranProgram,
    FortranSubroutine,
    FortranModule,
    NameSelector,
)
from ford._markdown import MetaMarkdown
import ford.sourceform

from itertools import chain

import pytest
from bs4 import BeautifulSoup


@pytest.fixture
def copy_fortran_file(tmp_path):
    # Reset unique name tracker
    setattr(ford.sourceform, "namelist", NameSelector())

    def copy_file(data) -> ProjectSettings:
        src_dir = tmp_path / "src"
        src_dir.mkdir()
        filename = src_dir / "test.f90"
        with open(filename, "w") as f:
            f.write(data)
        return ProjectSettings(src_dir=src_dir)

    return copy_file


def create_project(settings: ProjectSettings):
    project = Project(settings)
    project.correlate()
    return project


def test_use_without_rename(copy_fortran_file):
    data = """\
    module module1
      use module2, only: routine_1
      implicit none
      private
      interface routine
          module procedure :: routine_1
      end interface routine
      public :: routine
    end module module1

    module module2
      implicit none
      private
      public :: routine_1
    contains
      subroutine routine_1(i)
        integer,intent(in) :: i
        write(*,*) i
      end subroutine routine_1
    end module module2
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)
    assert set(project.modules[0].all_procs.keys()) == {"routine_1", "routine"}


def test_use_and_rename(copy_fortran_file):
    data = """\
    module module1
      use module2, only: routine_1 => routine
      implicit none
      private
      interface routine
          module procedure :: routine_1
      end interface routine
      public :: routine
    end module module1

    module module2
      implicit none
      private
      public :: routine
    contains
      subroutine routine(i)
        integer,intent(in) :: i
        write(*,*) i
      end subroutine routine
    end module module2
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)
    assert set(project.modules[0].all_procs.keys()) == {"routine_1", "routine"}


def test_module_use_only_everything(copy_fortran_file):
    data = """\
    module default_access
      ! No access keyword
      integer :: int_public, int_private
      private :: int_private
      real :: real_public
      real, private :: real_private

      type :: type_public
        complex :: component_public
        complex, private :: component_private
      end type type_public

      type :: type_private
        character(len=1) :: string_public
        character(len=1), private :: string_private
      end type type_private

      type type_public_no_attr
      end type type_public_no_attr

      type, public :: type_public_attr
      end type type_public_attr

      type, private :: type_private_attr
      end type type_private_attr

      private :: sub_private, func_private, type_private

    contains
      subroutine sub_public
      end subroutine sub_public

      subroutine sub_private
      end subroutine sub_private

      integer function func_public()
      end function func_public

      integer function func_private()
      end function func_private
    end module default_access

    module use_only_everything
      use default_access, only : int_public, real_public, type_public, sub_public, func_public
    end module use_only_everything
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)

    # Double-check we're looking at the right module
    assert project.modules[1].name == "use_only_everything"
    assert set(project.modules[1].all_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[1].pub_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[1].all_types.keys()) == {
        "type_public",
    }
    assert set(project.modules[1].pub_types.keys()) == {
        "type_public",
    }
    assert set(project.modules[1].all_vars.keys()) == {
        "int_public",
        "real_public",
    }
    assert set(project.modules[1].pub_vars.keys()) == {
        "int_public",
        "real_public",
    }


def test_module_use_only_everything_change_access(copy_fortran_file):
    data = """\
    module default_access
      ! No access keyword
      integer :: int_public, int_private
      private :: int_private
      real :: real_public
      real, private :: real_private

      type :: type_public
        complex :: component_public
        complex, private :: component_private
      end type type_public

      type :: type_private
        character(len=1) :: string_public
        character(len=1), private :: string_private
      end type type_private

      type type_public_no_attr
      end type type_public_no_attr

      type, public :: type_public_attr
      end type type_public_attr

      type, private :: type_private_attr
      end type type_private_attr

      private :: sub_private, func_private, type_private

    contains
      subroutine sub_public
      end subroutine sub_public

      subroutine sub_private
      end subroutine sub_private

      integer function func_public()
      end function func_public

      integer function func_private()
      end function func_private
    end module default_access

    module use_only_change_access
      use default_access, only : int_public, real_public, type_public, sub_public, func_public
      private
      public :: int_public, sub_public
    end module use_only_change_access
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)

    # Double-check we're looking at the right module
    assert project.modules[1].name == "use_only_change_access"
    assert set(project.modules[1].all_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[1].pub_procs.keys()) == {
        "sub_public",
    }
    assert set(project.modules[1].all_types.keys()) == {
        "type_public",
    }
    assert set(project.modules[1].pub_types.keys()) == set()
    assert set(project.modules[1].all_vars.keys()) == {
        "int_public",
        "real_public",
    }
    assert set(project.modules[1].pub_vars.keys()) == {
        "int_public",
    }


def test_module_use_everything(copy_fortran_file):
    data = """\
    module default_access
      ! No access keyword
      integer :: int_public, int_private
      private :: int_private
      real :: real_public
      real, private :: real_private

      type :: type_public
        complex :: component_public
        complex, private :: component_private
      end type type_public

      type :: type_private
        character(len=1) :: string_public
        character(len=1), private :: string_private
      end type type_private

      type type_no_attr
      end type type_no_attr

      type, public :: type_public_attr
      end type type_public_attr

      type, private :: type_private_attr
      end type type_private_attr

      private :: sub_private, func_private, type_private

    contains
      subroutine sub_public
      end subroutine sub_public

      subroutine sub_private
      end subroutine sub_private

      integer function func_public()
      end function func_public

      integer function func_private()
      end function func_private
    end module default_access

    module use_everything
      use default_access
    end module use_everything
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)

    # Double-check we're looking at the right module
    assert project.modules[1].name == "use_everything"
    assert set(project.modules[1].all_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[1].pub_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[1].all_types.keys()) == {
        "type_public",
        "type_no_attr",
        "type_public_attr",
    }
    assert set(project.modules[1].pub_types.keys()) == {
        "type_public",
        "type_no_attr",
        "type_public_attr",
    }
    assert set(project.modules[1].all_vars.keys()) == {
        "int_public",
        "real_public",
    }
    assert set(project.modules[1].pub_vars.keys()) == {
        "int_public",
        "real_public",
    }


def test_module_use_everything_reexport(copy_fortran_file):
    data = """\
    module default_access
      ! No access keyword
      integer :: int_public, int_private
      private :: int_private
      real :: real_public
      real, private :: real_private

      type :: type_public
        complex :: component_public
        complex, private :: component_private
      end type type_public

      type :: type_private
        character(len=1) :: string_public
        character(len=1), private :: string_private
      end type type_private

      type type_no_attr
      end type type_no_attr

      type, public :: type_public_attr
      end type type_public_attr

      type, private :: type_private_attr
      end type type_private_attr

      private :: sub_private, func_private, type_private

    contains
      subroutine sub_public
      end subroutine sub_public

      subroutine sub_private
      end subroutine sub_private

      integer function func_public()
      end function func_public

      integer function func_private()
      end function func_private
    end module default_access

    module use_everything
      use default_access
    end module use_everything

    module reexport
      use use_everything
    end module reexport
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)

    # Double-check we're looking at the right module
    assert project.modules[2].name == "reexport"
    assert set(project.modules[2].all_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[2].pub_procs.keys()) == {
        "sub_public",
        "func_public",
    }
    assert set(project.modules[2].all_types.keys()) == {
        "type_public",
        "type_no_attr",
        "type_public_attr",
    }
    assert set(project.modules[2].pub_types.keys()) == {
        "type_public",
        "type_no_attr",
        "type_public_attr",
    }
    assert set(project.modules[2].all_vars.keys()) == {
        "int_public",
        "real_public",
    }
    assert set(project.modules[2].pub_vars.keys()) == {
        "int_public",
        "real_public",
    }


def test_use_within_interface(copy_fortran_file):
    data = """\
    module module1
      implicit none
      private
      interface
        subroutine routine_1()
          use module2, only: routine_2
        end subroutine routine_1
      end interface
      public :: routine
    end module module1

    module module2
      implicit none
      private
      public :: routine_2
    contains
      subroutine routine_2(i)
        integer,intent(in) :: i
        write(*,*) i
      end subroutine routine_2
    end module module2
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)
    assert "routine_2" in project.modules[0].all_procs["routine_1"].procedure.all_procs


def test_member_in_other_module(copy_fortran_file):
    data = """\
    module module1
      use module2, only: testInterface
      type, public, abstract :: external_type
      contains
        procedure(testInterface), nopass, deferred :: testProc
      end type external_type
    end module module1

    module module2
      type, public, abstract :: internal_type
      contains
        procedure(testInterface), nopass, deferred :: testProc
      end type internal_type

      abstract interface
        pure function testInterface(input) result(output)
          real, intent(in) :: input
          real :: output
        end function testInterface
      end interface
    end module module2
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)

    module1 = project.modules[0]
    module2 = project.modules[1]

    assert (
        module1.all_types["external_type"].boundprocs[0].proto
        is module2.all_types["internal_type"].boundprocs[0].proto
    )

    assert (
        module2.all_types["internal_type"].boundprocs[0].proto.name == "testInterface"
    )


def test_display_derived_types(copy_fortran_file):
    """Check that we can set visibility attributes on types, #388"""

    data = """\
    module foo
      type no_attrs
      end type no_attrs

      type, public :: public_attr
      end type public_attr

      type, private :: private_attr
      end type private_attr
    end module foo
    """

    settings = copy_fortran_file(data)
    settings.display = ["public"]
    project = Project(settings)
    project.correlate()

    type_names = set([t.name for t in project.types])
    assert type_names == {"no_attrs", "public_attr"}


def test_display_derived_types_default_private(copy_fortran_file):
    """Check that we can set visibility attributes on types, #388"""

    data = """\
    module foo
      private

      type no_attrs
      end type no_attrs

      type, public :: public_attr
      end type public_attr

      type, private :: private_attr
      end type private_attr
    end module foo
    """

    settings = copy_fortran_file(data)
    settings.display = ["public"]
    project = create_project(settings)

    type_names = set([t.name for t in project.types])
    assert type_names == {"public_attr"}


def test_display_private_derived_types(copy_fortran_file):
    """Check that we can set visibility attributes on types, #388"""

    data = """\
    module foo
      private

      type no_attrs
      end type no_attrs

      type, public :: public_attr
      end type public_attr

      type, private :: private_attr
      end type private_attr
    end module foo
    """

    settings = copy_fortran_file(data)
    settings.display = ["public", "private"]
    project = create_project(settings)

    type_names = set([t.name for t in project.types])
    assert type_names == {"no_attrs", "public_attr", "private_attr"}


def test_display_private_module_procedure(copy_fortran_file):
    data = """\
      module foo
        interface
          module subroutine bar()
          end subroutine bar
        end interface
      end module foo

      submodule (foo) baz
      contains
        module procedure bar
        end procedure
      end submodule baz
      """

    settings = copy_fortran_file(data)
    settings.display = ["public", "protected", "private"]

    project = create_project(settings)
    assert project.submodprocedures[0].name == "bar"


def test_interface_type_name(copy_fortran_file):
    """Check for shared prototype list"""
    data = """\
    module foo
      type name_t
      end type name_t

      type(name_t) :: var, var2
    end module foo
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)
    proto_names = [var.proto[0].name for var in project.modules[0].variables]
    assert proto_names == ["name_t", "name_t"]


def test_imported_interface_type_name(copy_fortran_file):
    """Check for shared prototype list"""
    data = """\
    module foo
      type name_t
      end type name_t
    end module foo

    module bar
      use foo
      type(name_t), allocatable :: var(:), var2(:)
    end module bar
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)
    proto_names = [var.proto[0].name for var in project.modules[1].variables]
    assert proto_names == ["name_t", "name_t"]


def test_abstract_interface_type_name(copy_fortran_file):
    """Check for shared prototype list"""
    data = """\
    module foo
      abstract interface
        subroutine hello
        end subroutine hello
      end interface

      procedure(hello) :: proc1, proc2
    end module foo
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)
    proto_names = [var.proto[0].name for var in project.modules[0].variables]
    assert proto_names == ["hello", "hello"]


def test_imported_abstract_interface_type_name(copy_fortran_file):
    """Check for shared prototype list"""
    data = """\
    module foo
      abstract interface
        subroutine hello
        end subroutine hello
      end interface

    contains
      integer function goodbye()
        goodbye = 1
      end function goodbye
    end module foo

    module bar
      use foo
      procedure(hello) :: proc1, proc2
      procedure(goodbye) :: proc3, proc4
    end module bar
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)
    proto_names = [var.proto[0].name for var in project.modules[1].variables]
    assert proto_names == ["hello", "hello", "goodbye", "goodbye"]


def test_generic_interface_inherited_attrs(copy_fortran_file):
    data = """\
    module module1
      implicit none
      interface routine
        module subroutine routine_1(var)
          integer, dimension(:) :: var
        end subroutine routine_1
      end interface
      public :: routine
    end module module1

    submodule(module1) module2
      implicit none
    contains
      module procedure routine_1
      end procedure
    end submodule module2
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)

    module_implementation = project.modules[0].all_procs["routine_1"].module
    # Check that the module implementation has been linked
    assert not isinstance(module_implementation, bool)
    # Check that mod proc has the inherited arg
    assert isinstance(module_implementation.args[0], FortranVariable)


def test_display_internal_procedures(copy_fortran_file):
    """_very_ basic test of 'proc_internals' setting"""

    data = """\
    subroutine no_display_internals
    !! some docs
    integer :: local_variable
    !! This shouldn't be displayed
    end subroutine no_display_internals

    subroutine display_internals
    !! proc_internals: true
    !!
    !! some docs
    integer :: local_variable
    !! This shouldn't be displayed
    end subroutine display_internals
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)

    subroutine1 = project.procedures[0]
    subroutine2 = project.procedures[1]

    assert subroutine1.variables == []
    assert len(subroutine2.variables) == 1
    assert subroutine2.variables[0].name == "local_variable"


def test_exclude_dir(tmp_path):
    exclude_dir = tmp_path / "sub1" / "sub2"
    exclude_dir.mkdir(parents=True)
    src = tmp_path / "src"
    src.mkdir()

    with open(src / "include.f90", "w") as f:
        f.write("program foo\nend program")
    with open(exclude_dir / "exclude.f90", "w") as f:
        f.write("program bar\nend program")

    settings = ProjectSettings(
        src_dir=tmp_path, exclude_dir=normalise_path(tmp_path, "sub1")
    )
    project = Project(settings)

    program_names = {program.name for program in project.programs}
    assert program_names == {"foo"}


def test_exclude(tmp_path):
    exclude_dir = tmp_path / "sub1" / "sub2"
    exclude_dir.mkdir(parents=True)
    src = tmp_path / "src"
    src.mkdir()

    with open(src / "include.f90", "w") as f:
        f.write("program foo\nend program")
    with open(exclude_dir / "exclude.f90", "w") as f:
        f.write("program bar\nend program")

    settings = ProjectSettings(src_dir=tmp_path, exclude="exclude.f90")
    project = Project((settings))

    program_names = {program.name for program in project.programs}
    assert program_names == {"foo"}


def test_find_all_files(tmp_path):
    src = tmp_path / "src"
    src.mkdir(parents=True)
    sub1 = src / "sub1"
    sub1.mkdir()
    sub2 = src / "sub2"
    sub2.mkdir()
    (src / "file1.f90").touch()
    (src / "file2.F90").touch()
    (sub1 / "file3.f").touch()
    (sub1 / "not_this.f90").touch()
    (sub2 / "file4.c").touch()
    (sub2 / "not_this.f90").touch()

    exclude_dir = src / "bad_dir"
    exclude_dir.mkdir()
    exclude_sub = exclude_dir / "sub1"
    exclude_sub.mkdir()
    (exclude_dir / "file5.f90").touch()
    (exclude_sub / "file6.f90").touch()

    src2 = tmp_path / "src2"
    src2.mkdir()
    (src2 / "file7.f90").touch()

    settings = ProjectSettings(
        exclude=["src/sub1/not_this.f90"],
        exclude_dir=["src/bad_dir"],
        src_dir=[src, src2],
        extra_filetypes={"c": ExtraFileType("c", "//")},
    )
    settings.normalise_paths(tmp_path)

    files = sorted(find_all_files(settings))

    expected_files = sorted(
        [
            tmp_path / f
            for f in (
                "src/file1.f90",
                "src/file2.F90",
                "src/sub1/file3.f",
                "src/sub2/file4.c",
                "src/sub2/not_this.f90",
                "src2/file7.f90",
            )
        ]
    )

    assert files == expected_files


@pytest.mark.parametrize(
    "sort_kind, expected_order",
    [
        (
            "src",
            {
                "variables": ["b", "a", "d", "f", "c", "e", "g", "h", "j", "i"],
                "types": ["bet", "alef", "gimel"],
                "subroutines": ["bravo", "alpha", "charlie"],
                "functions": ["beaver", "aardvark", "cat"],
                "absinterfaces": ["apple", "carrot", "banana"],
            },
        ),
        (
            "alpha",
            {
                "variables": ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j"],
                "types": ["alef", "bet", "gimel"],
                "subroutines": ["alpha", "bravo", "charlie"],
                "functions": ["aardvark", "beaver", "cat"],
                "absinterfaces": ["apple", "banana", "carrot"],
            },
        ),
        (
            "permission",
            {
                "variables": ["d", "f", "c", "e", "b", "a", "g", "h", "j", "i"],
                "types": ["gimel", "alef", "bet"],
                "subroutines": ["charlie", "alpha", "bravo"],
                "functions": ["cat", "aardvark", "beaver"],
                "absinterfaces": ["apple", "carrot", "banana"],
            },
        ),
        (
            "permission-alpha",
            {
                "variables": ["d", "f", "c", "e", "a", "b", "g", "h", "i", "j"],
                "types": ["gimel", "alef", "bet"],
                "subroutines": ["charlie", "alpha", "bravo"],
                "functions": ["cat", "aardvark", "beaver"],
                "absinterfaces": ["apple", "banana", "carrot"],
            },
        ),
        (
            "type",
            {
                "variables": ["h", "g", "j", "i", "b", "a", "c", "e", "d", "f"],
                "types": ["bet", "alef", "gimel"],
                "subroutines": ["bravo", "alpha", "charlie"],
                "functions": ["cat", "beaver", "aardvark"],
                "absinterfaces": ["apple", "carrot", "banana"],
            },
        ),
        (
            "type-alpha",
            {
                "variables": ["h", "g", "i", "j", "a", "b", "c", "e", "d", "f"],
                "types": ["alef", "bet", "gimel"],
                "subroutines": ["alpha", "bravo", "charlie"],
                "functions": ["cat", "aardvark", "beaver"],
                "absinterfaces": ["apple", "banana", "carrot"],
            },
        ),
    ],
)
def test_sort(copy_fortran_file, sort_kind, expected_order):
    data = """\
    module mod_a
      type :: bet
      end type bet

      type :: alef
      end type alef

      type :: gimel
      end type gimel

      integer, private :: b, a
      type(gimel), public :: d, f
      type(alef), protected :: c, e

      character(len=2, kind=4) :: g
      character(len=1, kind=4) :: h
      character(len=1, kind=8) :: j, i
      private :: g, h, i, j

      abstract interface
        subroutine banana
        end subroutine banana

        function apple()
          import gimel
          class(gimel), allocatable :: apple
        end function apple

        type(alef) function carrot()
          import alef
        end function carrot
      end interface

      public :: gimel, charlie, cat
      protected :: alef, alpha, aardvark
      private :: bet, bravo, beaver

    contains
      subroutine bravo
      end subroutine bravo

      subroutine alpha
      end subroutine alpha

      subroutine charlie
      end subroutine charlie

      type(gimel) function beaver()
      end function beaver

      function aardvark()
        class(gimel), allocatable :: aardvark
      end function aardvark

      type(alef) function cat()
      end function cat
    end module mod_a
    """

    settings = copy_fortran_file(data)
    settings.sort = sort_kind
    settings.display = ["public", "private", "protected"]
    project = create_project(settings)
    module = project.modules[0]

    for entity_type in [
        "variables",
        "types",
        "subroutines",
        "functions",
        "absinterfaces",
    ]:
        entity_names = [entity.name for entity in getattr(module, entity_type)]
        assert entity_names == expected_order[entity_type], (sort_kind, entity_type)


def test_uses(copy_fortran_file):
    """Check that module `USE`s are matched up correctly"""

    data = """\
    module mod_a
    end module mod_a

    module mod_b
      use mod_a
    end module mod_b

    module mod_c
      use mod_a
      use mod_b
    end module mod_c

    module mod_d
      use external_module
    end module mod_d
    """

    settings = copy_fortran_file(data)
    settings.extra_mods = {"external_module": "http://example.com"}

    project = create_project(settings)
    mod_a = project.modules[0]
    mod_b = project.modules[1]
    mod_c = project.modules[2]
    mod_d = project.modules[3]

    assert mod_a.uses == set()
    assert mod_b.uses == {mod_a}
    assert mod_c.uses == {mod_a, mod_b}

    for link in mod_d.uses:
        assert link.get_url() == "http://example.com"


def test_submodule_uses(copy_fortran_file):
    """Check that module `USE`s are matched up correctly"""

    data = """\
    module mod_a
    end module mod_a

    submodule (mod_a) mod_b
    end submodule mod_b

    submodule (mod_a) mod_c
    end submodule mod_c

    submodule (mod_a:mod_c) mod_d
    end submodule mod_d
    """

    settings = copy_fortran_file(data)

    project = create_project(settings)
    mod_a = project.modules[0]
    mod_b = project.submodules[0]
    mod_c = project.submodules[1]
    mod_d = project.submodules[2]

    assert mod_a.uses == set()

    assert mod_b.parent_submodule is None
    assert mod_b.ancestor_module == mod_a

    assert mod_c.parent_submodule is None
    assert mod_c.ancestor_module == mod_a

    assert mod_d.parent_submodule == mod_c
    assert mod_d.ancestor_module == mod_a


def test_make_links(copy_fortran_file):
    links = "[[a]] [[b(type)]] [[b:c]] [[a:d]] [[b:e]] [[F(proc)]] [[A:G]] [[H]] [[I]]"

    data = f"""\
    module a !! {links}
      type b !! {links}
        integer :: c !! {links}
      contains
        final :: d !! {links}
        procedure :: e !! {links}
      end type b

      interface b !! {links}
        module procedure f
      end interface b

      type(b) :: g !! {links}
    contains
      subroutine d(self) !! {links}
        type(b) :: self !! {links}
      end subroutine d
      subroutine e(self) !! {links}
        class(b) :: self !! {links}
      end subroutine e
      function f() !! {links}
        type(b) :: f !! {links}
      end function f
    end module a

    program h !! {links}
    end program h

    submodule (a) i !! {links}
    end submodule i
    """
    settings = copy_fortran_file(data)
    project = create_project(settings)
    md = MetaMarkdown(project=project)
    project.markdown(md)

    expected_links = {
        "a": "../module/a.html",
        "b": "../type/b.html",
        "c": "../type/b.html#variable-c",
        "d": "../proc/d.html",
        "e": "../type/b.html#boundprocedure-e",
        "f": "../proc/f.html",
        "g": "../module/a.html#variable-g",
        "h": "../program/h.html",
        "i": "../module/i.html",
    }

    for item in chain(
        project.files[0].children,
        project.types[0].children,
        project.procedures[0].children,
        project.procedures[2].children,
    ):
        docstring = BeautifulSoup(item.doc, features="html.parser")
        link_locations = {a.string: a.get("href", None) for a in docstring("a")}

        assert link_locations == expected_links, (item, item.name)


def test_make_links_with_entity_spec(copy_fortran_file):
    links = (
        "[[a(module)]] [[b(type)]] [[b(type):c(variable)]] "
        "[[A(MODULE):D(SUBROUTINE)]] [[B(TYPE):E]] [[F(PROC)]] "
        "[[A(module):G(variable)]] [[H(program)]] [[I(SUBMODULE)]]"
    )

    data = f"""\
    module a !! {links}
      type b !! {links}
        integer :: c !! {links}
      contains
        final :: d !! {links}
        procedure :: e !! {links}
      end type b

      interface b !! {links}
        module procedure f
      end interface b

      type(b) :: g !! {links}
    contains
      subroutine d(self) !! {links}
        type(b) :: self !! {links}
      end subroutine d
      subroutine e(self) !! {links}
        class(b) :: self !! {links}
      end subroutine e
      function f() !! {links}
        type(b) :: f !! {links}
      end function f
    end module a

    program h !! {links}
    end program h

    submodule (a) i !! {links}
    end submodule i
    """
    settings = copy_fortran_file(data)
    project = create_project(settings)
    md = MetaMarkdown(project=project)
    project.markdown(md)

    expected_links = {
        "a": "../module/a.html",
        "b": "../type/b.html",
        "c": "../type/b.html#variable-c",
        "d": "../proc/d.html",
        "e": "../type/b.html#boundprocedure-e",
        "f": "../proc/f.html",
        "g": "../module/a.html#variable-g",
        "h": "../program/h.html",
        "i": "../module/i.html",
    }

    for item in chain(
        project.files[0].children,
        project.types[0].children,
        project.procedures[0].children,
        project.procedures[2].children,
    ):
        docstring = BeautifulSoup(item.doc, features="html.parser")
        link_locations = {a.string: a.get("href", None) for a in docstring("a")}

        assert link_locations == expected_links, (item, item.name)


def test_make_links_with_context(copy_fortran_file):
    data = """\
    module a !! [[g]]
      type b !! [[c]] [[e]] [[g]]
        integer :: c
      contains
        procedure :: e
      end type b

      type(b) :: g
    contains
      subroutine d(i) !! [[I]] [[F]] [[F:X]]
        type(b) :: i
        contains
          integer function f(x)
            integer :: x
          end function f
      end subroutine d
    end module a

    program h
    end program h
    """
    settings = copy_fortran_file(data)
    settings.proc_internals = True
    project = create_project(settings)
    md = MetaMarkdown(project=project)
    project.markdown(md)

    expected_links = {
        "a": "../module/a.html",
        "b": "../type/b.html",
        "c": "../type/b.html#variable-c",
        "d": "../proc/d.html",
        "e": "../type/b.html#boundprocedure-e",
        "f": "../proc/d.html#proc-f",
        "g": "../module/a.html#variable-g",
        "h": "../program/h.html",
        "i": "../proc/d.html#variable-i",
        "x": "../proc/d.html#variable-x",
    }

    for item in chain(
        project.files[0].children,
        project.modules[0].children,
        project.types[0].children,
        project.procedures[0].children,
    ):
        docstring = BeautifulSoup(item.doc, features="html.parser")

        for link in docstring("a"):
            assert link.string in expected_links, (item, item.name, link)
            expected_link = expected_links[link.string]

            assert link.get("href", None) == expected_link, (
                item,
                item.name,
                link.string,
            )


def test_submodule_procedure_issue_446(copy_fortran_file):
    data = """\
    module foo_mod
      interface
        module subroutine quaxx
        end subroutine quaxx
      end interface
    end module foo_mod

    submodule (foo_mod) bar_submod
    contains
      module procedure quaxx
        print*, "implementation"
      end procedure
    end submodule bar_submod
    """

    settings = copy_fortran_file(data)
    settings.display = ["public", "private", "protected"]
    project = create_project(settings)
    module = project.modules[0]

    interface = module.interfaces[0]
    assert interface.name == "quaxx"
    submodproc = module.descendants[0].modprocedures[0]

    assert interface.procedure.module == submodproc


def test_find_namelists(copy_fortran_file):
    data = """\
    module mod1
    contains
      subroutine sub1()
        namelist /namelist_b/ var_a
      end subroutine sub1
    end module mod1

    program prog
      integer :: var_b
    contains
      subroutine sub(var_b)
        integer :: var_d, var_a, var_b
        character(len=*), intent(in) :: string
        namelist /namelist_a/ var_a, var_b, var_c, var_d
        read(string, nml=namelist_a)
        write(*, *) "Read in:"
        write(*, nml=namelist_a)
      end subroutine sub
    end program prog
    """
    settings = copy_fortran_file(data)
    project = create_project(settings)
    assert len(project.namelists) == 2
    namelist_names = sorted((namelist.name for namelist in project.namelists))
    expected_names = sorted(("namelist_a", "namelist_b"))
    assert namelist_names == expected_names


def test_find(copy_fortran_file):
    data = """\
    module foo
      type :: test_type
        integer :: sub
      end type test_type
    contains
      subroutine sub
      end subroutine sub
    end module

    program foo
    end program
    """

    settings = copy_fortran_file(data)
    settings.display = ["public", "private"]
    project = create_project(settings)

    test_type = project.find("test_type")
    assert test_type.name == "test_type"

    subroutine_sub = project.find("sub")
    assert subroutine_sub.name == "sub"
    assert isinstance(subroutine_sub, FortranSubroutine)

    component_sub = project.find("test_type", child_name="sub")
    assert component_sub.name == "sub"
    assert isinstance(component_sub, FortranVariable)

    foo = project.find("foo")
    assert foo
    prog_foo = project.find("foo", entity="program")
    assert isinstance(prog_foo, FortranProgram)
    mod_foo = project.find("foo", entity="module")
    assert isinstance(mod_foo, FortranModule)


def test_links_in_deferred_bound_methods(copy_fortran_file):
    """Test for issue #591"""

    data = """\
    module test
        type, abstract :: foo_t
            contains
            !> See [[foo_t]] for more information about [[foo]]
            procedure(foo_iface), deferred :: foo
        end type
        abstract interface
            subroutine foo_iface(self)
                import :: foo_t
                class(foo_t), intent(in) :: self
            end subroutine
        end interface
    end module
    """

    settings = copy_fortran_file(data)
    project = create_project(settings)
    md = MetaMarkdown(settings.md_base_dir, project=project)
    project.markdown(md)

    docstring = BeautifulSoup(
        project.modules[0].types[0].boundprocs[0].doc, features="html.parser"
    )
    links = {link.text: link["href"] for link in docstring.find_all("a")}
    assert links["foo_t"].endswith("foo_t.html")
    assert links["foo"].endswith("foo_t.html#boundprocedure-foo")


def test_hide_undoc(copy_fortran_file):
    data = """\
    module foo
    contains
      !> A subroutine with a docstring
      subroutine with_doc
      end subroutine with_doc

      subroutine no_doc
      end subroutine no_doc
    end module
    """

    settings = copy_fortran_file(data)
    settings.hide_undoc = True
    project = create_project(settings)

    assert len(project.modules[0].subroutines) == 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  reader.py
#  This file is part of FORD.
#
#  Copyright 2014 Christopher MacMackin <cmacmackin@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#

import sys
import re
import ford.utils
import subprocess
from typing import List, Optional, Union, Sequence
from io import StringIO, TextIOWrapper

import os.path

from ford.console import warn
from ford.fixed2free2 import convertToFree
from ford._typing import PathLike


def _contains_unterminated_string(string: str) -> bool:
    """Return True if `string` contains an unterminated quote"""
    in_quote = False
    current_quote = None
    previous_char = None
    for char in string:
        # Non-quote characters don't bother us
        if char not in ("'", '"'):
            previous_char = char
            continue
        # Doubling-up a quote character doesn't make us leave a quote
        if char == previous_char:
            previous_char = char
            continue
        # If the current character is the same as the starting quote character
        # then we have left the quote (as we've dealt with doubled-quotes)
        if char == current_quote:
            in_quote = False
            current_quote = None
            previous_char = char
            continue
        # If we weren't in a quote before, we are now
        if not in_quote:
            current_quote = char
            in_quote = True
        previous_char = char
    return in_quote


def _compile_docmark(docmark: str) -> Optional[re.Pattern]:
    """Compile ``docmark`` into a regex pattern to match Fortran comments"""
    if not docmark:
        return None

    return re.compile(f"^([^\"'!]|('[^']*')|(\"[^\"]*\"))*(!{re.escape(docmark)}.*)$")


def _match_docmark(
    docmark: Optional[re.Pattern], line: str, in_quote: bool
) -> Optional[re.Match]:
    """If docmark exists, and we're not in a string literal, try to match it"""
    if in_quote:
        return None
    if not docmark:
        return None
    return docmark.match(line)


class FortranReader:
    """
    An iterator which will convert a free-form Fortran source file into
    a format more conducive for analyzing. It does the following:

    - combine line continuations into one
    - remove any normal comments and any comments following an ampersand
      (line continuation)
    - if there are documentation comments preceding a piece of code, buffer
      them and return them after the code, but before any documentation
      following it
    - keep any documentation comments and, if they are at the end of a line
      of actual code, place them on a new line
    - removes blank lines and trailing white-space
    - split lines along semicolons
    """

    # Regexes
    COM_RE = re.compile("^([^\"'!]|('[^']*')|(\"[^\"]*\"))*(!.*)$")
    SC_RE = re.compile("^([^;]*);(.*)$")

    def __init__(
        self,
        filename: str,
        docmark: str = "!",
        predocmark: str = "",
        docmark_alt: str = "",
        predocmark_alt: str = "",
        fixed: bool = False,
        length_limit: bool = True,
        preprocessor: Optional[List[str]] = None,
        macros: Optional[List[str]] = None,
        inc_dirs: Optional[Sequence[PathLike]] = None,
        encoding: str = "utf-8",
    ):
        # Check that none of the docmarks are the same
        if docmark == predocmark != "":
            raise ValueError("Error: docmark and predocmark are the same.")
        if docmark == docmark_alt != "":
            raise ValueError("Error: docmark and docmark_alt are the same.")
        if docmark == predocmark_alt != "":
            raise ValueError("Error: docmark and predocmark_alt are the same.")
        if docmark_alt == predocmark != "":
            raise ValueError("Error: docmark_alt and predocmark are the same.")
        if docmark_alt == predocmark_alt != "":
            raise ValueError("Error: docmark_alt and predocmark_alt are the same.")
        if predocmark == predocmark_alt != "":
            raise ValueError("Error: predocmark and predocmark_alt are the same.")

        self.name = filename
        self.fixed = fixed
        self.length_limit = length_limit
        self.inc_dirs = inc_dirs or []
        self.docbuffer: List[str] = []
        self.pending: List[str] = []
        self.prevdoc = False
        self.reading_alt = 0
        self.encoding = encoding

        self.docmark = docmark
        self.doc_re = _compile_docmark(docmark)
        self.predocmark = predocmark
        self.predoc_re = _compile_docmark(predocmark)
        self.docmark_alt = docmark_alt
        self.doc_alt_re = _compile_docmark(docmark_alt)
        self.predocmark_alt = predocmark_alt
        self.predoc_alt_re = _compile_docmark(predocmark_alt)

        self.line_number = 0

        self.reader: Union[StringIO, TextIOWrapper]

        if preprocessor:
            # Populate the macro definition and include directory path from
            # the input lists.  To define a macro we prepend '-D' and for an
            # include path we prepend '-I'.  It's important that we do not
            # prepend to an empty string as 'cpp ... -D file.F90' doesn't do
            # what is desired; use filter to remove these.
            macros = ["-D" + mac.strip() for mac in filter(None, macros or [])]
            incdirs = [f"-I{d}" for d in self.inc_dirs]
            preprocessor = preprocessor + macros + incdirs + [filename]
            command = " ".join(preprocessor)
            print(f"Preprocessing {filename}")
            try:
                out = subprocess.run(
                    preprocessor, encoding=encoding, check=True, capture_output=True
                )
                if out.stderr:
                    warn(
                        f"Warning when preprocessing {filename}:\n{command}\n{out.stderr}"
                    )
                self.reader = StringIO(out.stdout)
            except subprocess.CalledProcessError as err:
                warn(
                    f"error when preprocessing {filename}:\n{command}\n{err.stderr}\n"
                    "Reverting to unpreprocessed file"
                )
                self.reader = open(filename, "r", encoding=encoding)
        else:
            self.reader = open(filename, "r", encoding=encoding)

        if fixed:
            self.reader = convertToFree(self.reader, length_limit)

    def __iter__(self):
        return self

    def pass_back(self, line: str):
        self.pending.insert(0, line)

    def __next__(self):
        # If there are any lines waiting to be returned, return them
        if len(self.pending) != 0:
            self.include()
            self.prevdoc = False
            return self.pending.pop(0)
        # If there are any documentation lines waiting to be returned, return them.
        # This can happen for inline and preceding docs
        elif len(self.docbuffer) != 0:
            self.prevdoc = True
            return self.docbuffer.pop(0)

        # Loop through the source code until you have a complete line (including
        # all line continuations), or a complete preceding doc block
        done = False
        continued = False
        reading_predoc = False
        reading_predoc_alt = 0
        linebuffer = ""

        while not done:
            line = next(self.reader)

            self.line_number += 1

            in_quote = _contains_unterminated_string(linebuffer)

            if len(line.strip()) > 0 and line.strip()[0] == "#":
                continue

            # Capture any preceding documentation comments
            match = _match_docmark(self.predoc_re, line, in_quote)
            if match:
                # Switch to predoc: following comment lines are predoc until the end of the block
                reading_predoc = True
                self.reading_alt = 0
                reading_predoc_alt = 0
                # Substitute predocmark with docmark
                tmp = match.group(4)
                tmp = tmp[:1] + self.docmark + tmp[1 + len(self.predocmark) :]
                self.docbuffer.append(tmp)
                if len(line[0 : match.start(4)].strip()) > 0:
                    raise ValueError(
                        f"Preceding documentation lines can not be inline: {line}"
                    )

            # Check for alternate preceding documentation
            match = _match_docmark(self.predoc_alt_re, line, in_quote)
            if match:
                # Switch to doc_alt: following comment lines are documentation until end of the block
                reading_predoc_alt = 1
                self.reading_alt = 0
                reading_predoc = False
                # Substitute predocmark_alt with docmark
                tmp = match.group(4)
                tmp = tmp[:1] + self.docmark + tmp[1 + len(self.predocmark_alt) :]
                self.docbuffer.append(tmp)
                if len(line[0 : match.start(4)].strip()) > 0:
                    raise RuntimeError(
                        f"In file {self.name}\n{self.line_number}|  {line.strip()}\n"
                        f"{'^':>{len(str(self.line_number)) + match.start(4) + 4}}\n"
                        f"Preceding alternate documentation lines can not be inline\n"
                        f"Use the docmark '!{self.docmark}' instead"
                    )

            # Check for alternate succeeding documentation
            match = _match_docmark(self.doc_alt_re, line, in_quote)
            if match:
                # Switch to doc_alt: following comment lines are documentation until end of the block
                self.reading_alt = 1
                reading_predoc = False
                reading_predoc_alt = 0
                # Substitute predocmark_alt with docmark
                tmp = match.group(4)
                tmp = tmp[:1] + self.docmark + tmp[1 + len(self.docmark_alt) :]
                self.docbuffer.append(tmp)
                if len(line[0 : match.start(4)].strip()) > 0:
                    raise ValueError(
                        f"Alternate documentation lines can not be inline: {line}"
                    )

            # Capture any documentation comments
            match = _match_docmark(self.doc_re, line, in_quote)
            if match:
                self.reading_alt = 0
                reading_predoc_alt = 0
                self.docbuffer.append(match.group(4))
                line = line[0 : match.start(4)]

            if len(line.strip()) == 0 or line.strip()[0] != "!":
                self.reading_alt = 0

            if len(line.strip()) != 0 and line.strip()[0] != "!":
                reading_predoc_alt = 0

            # Remove any regular comments, unless following an alternative (pre)docmark
            match = _match_docmark(self.COM_RE, line, in_quote)
            if match:
                if (reading_predoc_alt > 1 or self.reading_alt > 1) and len(
                    line[0 : match.start(4)].strip()
                ) == 0:
                    tmp = match.group(4)
                    tmp = tmp[:1] + self.docmark + tmp[1:]
                    self.docbuffer.append(tmp)
                line = line[0 : match.start(4)]
            line = line.strip()

            # If this is a blank line following previous documentation, return
            # a line of empty documentation.
            if len(line) == 0:
                if self.prevdoc and len(self.docbuffer) == 0:
                    # ~ self.prevdoc = False
                    self.docbuffer.append("!" + self.docmark)
            else:
                reading_predoc = False
                reading_predoc_alt = 0
                self.reading_alt = 0
                # Check if line is immediate continuation of previous
                if line[0] == "&":
                    if continued:
                        line = line[1:]
                        if len(line.strip()) == 0:
                            # If the line contained only an "&" and `continued==True` then
                            # we keep going.
                            continue
                    elif len(line.strip()) == 1:
                        continue
                    else:
                        raise ValueError(
                            f"Can not start a new line in Fortran with '&': {line}"
                        )
                else:
                    linebuffer = linebuffer.strip() + " "
                # Check if line will be continued
                if line[-1] == "&":
                    continued = True
                    line = line[0:-1]
                else:
                    continued = False

            if self.reading_alt > 0:
                self.reading_alt += 1

            if reading_predoc_alt > 0:
                reading_predoc_alt += 1

            # Add this line to the buffer then check whether we're done here
            linebuffer += line
            # ~ print(((len(self.docbuffer) > 0) or (len(linebuffer) > 0)), not continued, not reading_predoc, (reading_predoc_alt == 0))
            done = (
                ((len(self.docbuffer) > 0) or (len(linebuffer) > 0))
                and not continued
                and not reading_predoc
                and (reading_predoc_alt == 0)
            )

        # Split buffer with semicolons
        # ~ print(count,linebuffer,len(linebuffer))
        frags = ford.utils.quote_split(";", linebuffer)
        self.pending.extend([s.strip() for s in frags if len(s) > 0])

        # Return the line
        if len(self.pending) > 0:
            self.include()
            self.prevdoc = False
            return self.pending.pop(0)
        else:
            tmp = self.docbuffer.pop(0)
            if tmp != "!" + self.docmark:
                self.prevdoc = True
            return tmp

    def include(self):
        """
        If the next line is an include statement, inserts the contents
        of the included file into the pending buffer.
        """
        if len(self.pending) == 0 or not self.pending[0].lower().startswith("include "):
            return
        curpending = self.pending.pop(0)
        name = curpending[8:].strip()[1:-1]
        for b in [os.path.dirname(self.name)] + self.inc_dirs:
            pname = os.path.abspath(os.path.expanduser(os.path.join(b, name)))
            if os.path.isfile(pname):
                name = pname
                break
        else:
            msg = f'Can not find include file "{name}"'
            if name.endswith(".h"):
                warn(msg)
                # undo pop and return
                self.pending = [curpending] + self.pending
                return
            raise FileNotFoundError(msg)
        self.pending = (
            list(
                FortranReader(
                    name,
                    self.docmark,
                    self.predocmark,
                    self.docmark_alt,
                    self.predocmark_alt,
                    self.fixed,
                    self.length_limit,
                    inc_dirs=self.inc_dirs,
                    encoding=self.encoding,
                )
            )
            + self.pending
        )


if __name__ == "__main__":
    filename = sys.argv[1]
    for line in FortranReader(
        filename,
        docmark="!",
        predocmark=">",
        docmark_alt="#",
        predocmark_alt="<",
    ):
        print(line)



import ford
from textwrap import dedent
from pathlib import Path
import sys
import pytest

import tomli_w

from conftest import gfortran_is_not_installed


class FakeFile:
    name = "test file"


def test_quiet_false():
    _, data = ford.load_settings("quiet: False")
    assert data.quiet is False
    _, data2 = ford.load_settings("quiet: True")
    assert data2.quiet is True


def test_toml(tmp_path):
    settings_file = tmp_path / "fpm.toml"
    settings = {
        "extra": {
            "ford": {
                "quiet": True,
                "display": ["public", "protected"],
                "src_dir": "./source",
            }
        }
    }
    settings_file.write_text(tomli_w.dumps(settings))

    _, data = ford.load_settings("", tmp_path)

    assert data.quiet is True
    assert data.display[0] == "public"
    assert data.display[1] == "protected"
    assert data.src_dir == ["./source"]


def test_quiet_command_line():
    """Check that setting --quiet on the command line overrides project file"""

    data, _ = ford.parse_arguments(
        {"quiet": True, "preprocess": False}, "", ford.ProjectSettings(quiet=False)
    )
    assert data.quiet is True
    data, _ = ford.parse_arguments(
        {"quiet": False, "preprocess": False}, "", (ford.ProjectSettings(quiet=True))
    )
    assert data.quiet is False


def test_list_input():
    """Check that setting a non-list option is turned into a single string"""

    settings = """\
    include: line1
             line2
    summary: This
             is
             one
             string
    """
    _, data = ford.load_settings(dedent(settings))

    assert len(data.include) == 2
    assert data.summary == "This\nis\none\nstring"


def test_path_normalisation():
    """Check that paths get normalised correctly"""

    data, _ = ford.parse_arguments(
        {"preprocess": False},
        "",
        ford.ProjectSettings(page_dir="my_pages", src_dir=["src1", "src2"]),
        directory="/prefix/path",
    )
    assert data.page_dir == Path("/prefix/path/my_pages").absolute()
    assert data.src_dir == [
        Path("/prefix/path/src1").absolute(),
        Path("/prefix/path/src2").absolute(),
    ]


def test_source_not_subdir_output():
    """Check if the src_dir is correctly detected as being a subdirectory of output_dir"""

    # This should be fine
    data, _ = ford.parse_arguments(
        {"src_dir": ["/1/2/3", "4/5"], "output_dir": "/3/4", "preprocess": False},
        "",
        ford.ProjectSettings(),
        directory="/prefix",
    )

    # This shouldn't be
    with pytest.raises(ValueError):
        data, _ = ford.parse_arguments(
            {"src_dir": ["4/5", "/1/2/3"], "output_dir": "/1/2", "preprocess": False},
            "",
            ford.ProjectSettings(),
            directory="/prefix",
        )
    # src_dir == output_dir
    with pytest.raises(ValueError):
        data, _ = ford.parse_arguments(
            {"src_dir": ["/1/2/"], "output_dir": "/1/2", "preprocess": False},
            "",
            ford.ProjectSettings(),
            directory="/prefix",
        )


def test_repeated_docmark():
    """Check that setting --quiet on the command line overrides project file"""

    with pytest.raises(ValueError):
        ford.parse_arguments(
            {"preprocess": False},
            "",
            (ford.ProjectSettings(**{"docmark": "!", "predocmark": "!"})),
        )

    with pytest.raises(ValueError):
        ford.parse_arguments(
            {"preprocess": False},
            "",
            (ford.ProjectSettings(**{"docmark": "!<", "predocmark_alt": "!<"})),
        )

    with pytest.raises(ValueError):
        ford.parse_arguments(
            {"preprocess": False},
            "",
            (ford.ProjectSettings(**{"docmark_alt": "!!", "predocmark_alt": "!!"})),
        )


def test_no_preprocessor():
    data, _ = ford.parse_arguments(
        {}, "", (ford.ProjectSettings(**{"preprocess": False}))
    )

    assert data.fpp_extensions == []


def test_bad_preprocessor():
    with pytest.raises(SystemExit):
        ford.parse_arguments(
            {"project_file": FakeFile()},
            "",
            (ford.ProjectSettings(**{"preprocessor": "false"})),
        )


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="FIXME: Need portable do-nothing command"
)
def test_maybe_ok_preprocessor():
    data, _ = ford.parse_arguments(
        {}, "", (ford.ProjectSettings(**{"preprocessor": "true"}))
    )

    if data.preprocess is True:
        assert data.preprocessor == "true"


@pytest.mark.skipif(
    gfortran_is_not_installed(), reason="Requires gfortran to be installed"
)
def test_gfortran_preprocessor():
    data, _ = ford.parse_arguments(
        {}, "", (ford.ProjectSettings(**{"preprocessor": "gfortran -E"}))
    )

    assert data.preprocess is True


def test_absolute_src_dir(monkeypatch, tmp_path):
    project_file = tmp_path / "example.md"
    project_file.write_text("preprocess: False")
    src_dir = tmp_path / "not_here"

    with monkeypatch.context() as m:
        m.setattr(sys, "argv", ["ford", str(project_file)])
        args, _ = ford.initialize()

    assert args.src_dir == [tmp_path / "./src"]

    with monkeypatch.context() as m:
        m.setattr(sys, "argv", ["ford", str(project_file), "--src_dir", str(src_dir)])
        args, _ = ford.initialize()

    assert args.src_dir == [src_dir]

    with monkeypatch.context() as m:
        m.setattr(
            sys, "argv", ["ford", "--src_dir", str(src_dir), "--", str(project_file)]
        )
        args, _ = ford.initialize()

    assert args.src_dir == [src_dir]


def test_output_dir_cli(monkeypatch, tmp_path):
    project_file = tmp_path / "example.md"
    project_file.write_text("preprocess: False")

    with monkeypatch.context() as m:
        m.setattr(sys, "argv", ["ford", str(project_file), "--output_dir", "something"])
        settings, _ = ford.initialize()

    assert settings.output_dir == tmp_path / "something"

    with open(project_file, "a") as f:
        f.write("\noutput_dir: something_else")

    with monkeypatch.context() as m:
        m.setattr(sys, "argv", ["ford", str(project_file)])
        settings, _ = ford.initialize()

    assert settings.output_dir == tmp_path / "something_else"


def test_config_option():
    command_line_args = {
        "preprocess": False,
        "config": "quiet = true; display = ['public']; alias = {a = 'b', c = 'd'}",
    }

    data, _ = ford.parse_arguments(
        command_line_args, "", ford.ProjectSettings(quiet=False)
    )

    assert data.quiet is True
    assert data.display == ["public"]
    assert data.alias == {"a": "b", "c": "d"}


def test_external_links_command_line_arg():
    data, _ = ford.parse_arguments(
        {"external": "remote = https://some.link"}, "", ford.ProjectSettings()
    )

    assert data.external == {"remote": "https://some.link"}
